package ru.dbotthepony.mc.otm.client.screen

import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.client.gui.screens.inventory.InventoryScreen
import ru.dbotthepony.mc.otm.client.mousePos
import ru.dbotthepony.mc.otm.client.moveMousePosScaled
import ru.dbotthepony.mc.otm.client.render.Widgets18
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.client.render.sprite
import ru.dbotthepony.mc.otm.client.screen.panels.*
import ru.dbotthepony.mc.otm.client.screen.panels.button.LargeRectangleButtonPanel
import ru.dbotthepony.mc.otm.client.screen.panels.slot.AbstractSlotPanel
import ru.dbotthepony.mc.otm.client.screen.panels.slot.UserFilteredSlotPanel
import ru.dbotthepony.mc.otm.client.screen.panels.slot.SlotPanel
import ru.dbotthepony.mc.otm.client.screen.panels.util.DiscreteScrollBarPanel
import ru.dbotthepony.mc.otm.client.setMousePos
import ru.dbotthepony.mc.otm.client.shouldOpenVanillaInventory
import ru.dbotthepony.mc.otm.core.math.integerDivisionDown
import ru.dbotthepony.mc.otm.menu.ExoPackInventoryMenu
import ru.dbotthepony.mc.otm.network.ExoPackMenuOpen
import ru.dbotthepony.mc.otm.network.MatteryPlayerNetworkChannel
import yalter.mousetweaks.api.MouseTweaksDisableWheelTweak

@MouseTweaksDisableWheelTweak
class ExoPackInventoryScreen(menu: ExoPackInventoryMenu) : MatteryScreen<ExoPackInventoryMenu>(menu, TranslatableComponent("otm.gui.exopack")) {
	private lateinit var mainInventoryLine: EditablePanel<ExoPackInventoryScreen>
	private lateinit var scrollPanel: DiscreteScrollBarPanel<ExoPackInventoryScreen>

	override fun updateInventoryRows(old: Int) {
		mainFrame!!.height = FRAME_BASE_HEIGHT + inventoryRows * AbstractSlotPanel.SIZE
		mainInventoryLine.height = AbstractSlotPanel.SIZE * inventoryRows

		for (i in scrollPanel.scroll + inventoryRows until scrollPanel.scroll + old) {
			getInventorySlotsRow(i).visible = false
		}

		for (i in scrollPanel.scroll until scrollPanel.scroll + inventoryRows) {
			getInventorySlotsRow(i).also {
				it.parent = mainInventoryLine
				it.y = AbstractSlotPanel.SIZE * (i - scrollPanel.scroll)
				it.visible = true
			}
		}

		scrollPanel.scroll = scrollPanel.scroll
	}

	override fun makeMainFrame(): FramePanel<MatteryScreen<*>> {
		val frame = FramePanel(this, width = 200f, height = FRAME_BASE_HEIGHT + inventoryRows * AbstractSlotPanel.SIZE, title = this.title)

		val hotbarStrip = EditablePanel(this, frame, height = 18f)
		hotbarStrip.dock = Dock.BOTTOM

		hotbarStrip.setDockMargin(top = 3f)

		scrollPanel = DiscreteScrollBarPanel(this, null, maxScroll = { integerDivisionDown(menu.playerCombinedInventorySlots.size, 9) },
			scrollCallback = {
					_, old, new ->

				for (i in old until old + inventoryRows) {
					getInventorySlotsRow(i).visible = false
				}

				for (i in new until new + inventoryRows) {
					val row = getInventorySlotsRow(i)
					row.visible = true
					row.y = (i - new) * AbstractSlotPanel.SIZE
					row.parent = mainInventoryLine
				}

				lastScroll = new
			})

		mainInventoryLine = object : EditablePanel<ExoPackInventoryScreen>(this@ExoPackInventoryScreen, frame, height = AbstractSlotPanel.SIZE * inventoryRows) {
			override fun mouseScrolledInner(x: Double, y: Double, scroll: Double): Boolean {
				return scrollPanel.mouseScrolledInner(x, y, scroll)
			}
		}

		scrollPanel.scroll = lastScroll
		scrollPanel.parent = mainInventoryLine

		mainInventoryLine.dock = Dock.BOTTOM

		for (slot in menu.playerHotbarSlots) {
			UserFilteredSlotPanel.of(this, hotbarStrip, slot).also {
				it.dock = Dock.LEFT
			}
		}

		val offhand = SlotPanel(this, hotbarStrip, menu.offhandSlot)
		offhand.dock = Dock.RIGHT

		for (i in scrollPanel.scroll until scrollPanel.scroll + inventoryRows) {
			getInventorySlotsRow(i).also {
				it.parent = mainInventoryLine
				it.y = AbstractSlotPanel.SIZE * (i - scrollPanel.scroll)
			}
		}

		val topLine = EditablePanel(this, frame, height = PlayerEquipmentPanel.HEIGHT)
		topLine.dock = Dock.TOP
		topLine.setDockMargin(top = 3f)

		val equipment = PlayerEquipmentPanel(this, topLine, armorSlots = menu.armorSlots)
		equipment.dock = Dock.LEFT
		equipment.setDockMargin(right = 3f)

		val craftingCanvas = EditablePanel(this, topLine, width = 90f)
		craftingCanvas.dock = Dock.LEFT

		val craftingSlotsCanvas = EditablePanel(this, craftingCanvas)

		if (menu.capability.isExoPackCraftingUpgraded) {
			craftingSlotsCanvas.setSize(54f, 54f)

			for (row in 0 .. 2) {
				for (column in 0 .. 2) {
					SlotPanel(this, craftingSlotsCanvas, menu.craftingSlots[row * 3 + column], x = column * 18f, y = row * 18f)
				}
			}
		} else {
			craftingSlotsCanvas.setSize(36f, 36f)

			SlotPanel(this, craftingSlotsCanvas, menu.craftingSlots[0])
			SlotPanel(this, craftingSlotsCanvas, menu.craftingSlots[1], x = 18f)
			SlotPanel(this, craftingSlotsCanvas, menu.craftingSlots[2], y = 18f)
			SlotPanel(this, craftingSlotsCanvas, menu.craftingSlots[3], x = 18f, y = 18f)
		}

		craftingSlotsCanvas.y = topLine.height / 2f - craftingSlotsCanvas.height / 2f

		val resultPanel = SlotPanel(this, craftingCanvas, menu.craftingResultSlot, x = craftingCanvas.width - 18f, y = topLine.height / 2f - 9f)

		object : EditablePanel<ExoPackInventoryScreen>(
			this@ExoPackInventoryScreen,
			craftingCanvas,
			x = craftingSlotsCanvas.width,
			width = craftingCanvas.width - resultPanel.width - craftingSlotsCanvas.width,
			height = topLine.height
		) {
			override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
				CRAFT_ARROW.render(stack, x = width / 2f - CRAFT_ARROW.width / 2f, y = height / 2f - CRAFT_ARROW.height / 2f)
			}
		}

		scrollPanel.dock = Dock.RIGHT
		scrollPanel.setDockMargin(right = 3f)

		LargeRectangleButtonPanel(this, frame, x = frame.width - 2f - LargeRectangleButtonPanel.SIZE, y = -LargeRectangleButtonPanel.SIZE - 2f, skinElement = Widgets18.RETURN_ARROW_LEFT, onPress = {
			shouldOpenVanillaInventory = true
			val minecraft = minecraft!!

			val (mouseX, mouseY) = mousePos

			onClose()
			minecraft.setScreen(InventoryScreen(minecraft.player!!))

			setMousePos(mouseX, mouseY)
		}).also { it.tooltip = TranslatableComponent("otm.gui.exopack.go_back") }

		makeInventoryRowsControls(frame, frame.width + 2f, frame.height.coerceAtMost(95f)) { movePixels ->
			frame.y += movePixels
			moveMousePosScaled(y = movePixels)
		}

		var x = -4f

		if (menu.curiosSlots.isNotEmpty()) {
			val curios = makeCuriosPanel(this, frame, menu.curiosSlots)!!

			x -= curios.width
			curios.x = x
		}

		EffectListPanel(this, frame, menu.ply, x - 56f, gridHeight = 6).tick()

		return frame
	}

	init {
		MatteryPlayerNetworkChannel.sendToServer(ExoPackMenuOpen)
		ru.dbotthepony.mc.otm.client.minecraft.player?.containerMenu = menu
	}

	companion object {
		val ENTITY_RECTANGLE = INVENTORY_LOCATION.sprite(25f, 7f, 51f, 72f)
		val CRAFT_ARROW = INVENTORY_LOCATION.sprite(135f, 29f, 16f, 13f)

		const val FRAME_BASE_HEIGHT = 126f
	}
}
